/**
 * Copyright 2008 Bluestem Software LLC.  All Rights Reserved.
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

package org.bluestemsoftware.open.eoa.ext.binding.soap.dfault.http;

import java.util.Map;
import java.util.Set;

import javax.xml.namespace.QName;

import org.apache.axiom.om.OMOutputFormat;
import org.apache.axiom.soap.SOAPEnvelope;
import org.apache.axiom.soap.SOAPFault;
import org.apache.axiom.soap.SOAPFaultCode;
import org.bluestemsoftware.open.eoa.ext.binding.soap.dfault.util.Constants;
import org.bluestemsoftware.specification.eoa.component.binding.rt.BindingFaultRT;
import org.bluestemsoftware.specification.eoa.ext.binding.soap.rt.SOAP12HTTPBindingRT;
import org.bluestemsoftware.specification.eoa.ext.binding.soap.rt.SOAPBindingConfiguration;
import org.bluestemsoftware.specification.eoa.ext.binding.soap.rt.SOAPSenderFault;
import org.bluestemsoftware.specification.eoa.ext.feature.http.server.HTTPServerRequest;

public final class SOAP12HTTPBindingImpl extends SOAPHTTPBindingImpl implements SOAP12HTTPBindingRT.Provider {

    public SOAP12HTTPBindingImpl(SOAPBindingConfiguration configuration, Set<String> features) {
        super(configuration, features);
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.open.eoa.ext.binding.soap.dfault.http.SOAPHTTPProtocolImpl#getContentType(org.apache.axiom.om.OMOutputFormat,
     *      java.lang.String)
     */
    @Override
    public String getContentType(OMOutputFormat omFormat, String soapAction) {
        String contentType = omFormat.getContentType();
        String encoding = omFormat.getCharSetEncoding();
        if (encoding != null) {
            contentType = contentType + "; charset=" + encoding;
        }
        if (soapAction != null && soapAction.trim().length() > 1) {
            contentType = contentType + "; action=\"" + soapAction + "\"";
        }
        return contentType;
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.open.eoa.ext.binding.soap.dfault.http.SOAPHTTPProtocolImpl#getSOAPAction(HTTPServerRequest)
     */
    protected String getSOAPAction(HTTPServerRequest request) {

        // parses optional soap action parameter from content type, e.g.
        // application/soap+xml;action="http://foo"[;charset=utf-8]

        String contentType = request.getContentType();
        int index = contentType.indexOf("action");
        if (index > -1) {
            String temp = contentType.substring(index, contentType.length());
            int start = temp.indexOf("=") + 1;
            int end = temp.indexOf(";");
            end = end > -1 ? end : temp.length();
            temp = temp.substring(start + 1, end - 1).trim();
            return temp.equals("") ? null : temp;
        } else {
            return null;
        }

    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.open.eoa.ext.binding.soap.dfault.http.SOAPHTTPBindingImpl#getMediaType(java.lang.String)
     */
    @Override
    protected String getMediaType(String contentType) throws BindingFaultRT {
        String mediaType = null;
        int index = contentType.indexOf(';');
        if (index > 0) {
            mediaType = contentType.substring(0, index);
        } else {
            mediaType = contentType;
        }
        if (mediaType.equals(Constants.SOAP_12_MEDIA_TYPE) || mediaType.equals(Constants.SOAP_MTOM_MEDIA_TYPE)) {
            return mediaType;
        } else {
            throw new SOAPSenderFault(getSOAPVersion(), "Found media type '"
                    + mediaType
                    + "'. Expected "
                    + Constants.SOAP_12_MEDIA_TYPE);
        }
    }
    
    @Override
    protected boolean isSOAPProtocolFault(SOAPEnvelope soapEnvelope) {
        if (soapEnvelope == null) {
            return false;
        }
        if (soapEnvelope.getBody().hasFault()) {
            SOAPFault sf = soapEnvelope.getBody().getFault();
            SOAPFaultCode code = sf.getCode();
            QName name = code.getTextAsQName();
            if (name.equals(Constants.SOAP_12_MUST_UNDERSTAND_FAULT)) {
                return true;
            } else if (name.equals(Constants.SOAP_12_VERSION_MISMATCH_FAULT)) {
                return true;
            } else {
                return false;
            }
        } else {
            return false;
        }
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.open.eoa.ext.binding.soap.dfault.http.SOAPHTTPBindingImpl#setSOAPAction(java.util.Map)
     */
    @Override
    protected String setSOAPAction(Map<String, String> requestHeaders) {

        // soapAction header is optional for soap 1.2. to avoid action
        // mismatch faults, we will always leave soap action undefined
        // on request, i.e. receiver should use wsa action to dispatch
        
        return null;

    }

}
