/**
 * Copyright 2008 Bluestem Software LLC.  All Rights Reserved.
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

package org.bluestemsoftware.open.eoa.ext.binding.soap.dfault.util.dao;

import java.sql.SQLException;

import javax.sql.DataSource;

import org.bluestemsoftware.open.eoa.ext.binding.soap.dfault.util.dao.derby.DerbyMessageModuleDAO;
import org.bluestemsoftware.specification.eoa.component.intrface.rt.MessageModule;
import org.bluestemsoftware.specification.eoa.component.intrface.rt.ActionContext.MessageModules;
import org.bluestemsoftware.specification.eoa.ext.binding.soap.SOAPBindingException;
import org.bluestemsoftware.specification.eoa.ext.connector.db.DataSourceConnector;
import org.bluestemsoftware.specification.eoa.ext.connector.db.DataSourceConnector.Dialect;

public abstract class MessageModuleDAO {
    
    public abstract MessageModuleDAO init() throws SOAPBindingException;
    
    public abstract void destroy();    
    
    /**
     * 
     * @param messageID
     * @param myService
     *        required in addition to messageid to make primary key unique, i.e. when partner
     *        engine is defined on same machine. true if receiving my service request or
     *        sending my service response. false if sending partner service request or
     *        receiving partner service response.
     * @param messageModule
     * @throws SQLException
     * @throws DAOException
     */
    public abstract void insertMessageModule(String messageID, boolean myService, MessageModule messageModule) throws SQLException, DAOException ;
	
    /**
     * 
     * @param messageID
     * @param myService
     *        required in addition to messageid to make primary key unique, i.e. when partner
     *        engine is defined on same machine. true if receiving my service request or
     *        sending my service response. false if sending partner service request or
     *        receiving partner service response.
     * @return
     * @throws SQLException
     * @throws DAOException
     */
    public abstract MessageModules selectMessageModules(String messageID, boolean myService) throws SQLException, DAOException ;
    
    /**
     * 
     * @param messageID
     * @param myService
     *        required in addition to messageid to make primary key unique, i.e. when partner
     *        engine is defined on same machine. true if receiving my service request or
     *        sending my service response. false if sending partner service request or
     *        receiving partner service response. 
     * @throws SQLException
     */
    public abstract void deleteMessageModules(String messageID, boolean myService) throws SQLException ;
    
    /**
     * 
     * @param messageID
     * @param myService
     *        required in addition to messageid to make primary key unique, i.e. when partner
     *        engine is defined on same machine. true if receiving my service request or
     *        sending my service response. false if sending partner service request or
     *        receiving partner service response.
     * @param moduleType
     * @throws SQLException
     */
    public abstract void deleteMessageModule(String messageID, boolean myService, String moduleType) throws SQLException ;
    
    public static MessageModuleDAO getInstance(DataSource dataSource) throws SOAPBindingException {
        Dialect dialect = ((DataSourceConnector)dataSource).getDataSourceDialect();
        if (dialect == Dialect.DERBY) {
            return DerbyMessageModuleDAO.getInstance(dataSource);
        } else {
            throw new SOAPBindingException("Dialect " + dialect + " not currently supported.");
        }
    }
	
}