/**
 * Copyright 2008 Bluestem Software LLC.  All Rights Reserved.
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

package org.bluestemsoftware.open.eoa.ext.binding.soap.dfault.vmtp;

import java.io.File;
import java.io.FileOutputStream;
import java.io.OutputStream;
import java.sql.SQLException;
import java.sql.SQLIntegrityConstraintViolationException;
import java.util.Arrays;
import java.util.List;
import java.util.Map;
import java.util.Set;

import javax.jms.DeliveryMode;
import javax.xml.namespace.QName;
import javax.xml.stream.XMLStreamWriter;

import org.apache.axiom.om.OMElement;
import org.apache.axiom.soap.SOAPEnvelope;
import org.apache.commons.io.IOUtils;
import org.bluestemsoftware.open.eoa.aspect.axiom.util.STAXUtils;
import org.bluestemsoftware.open.eoa.ext.binding.soap.dfault.SOAPBindingImpl;
import org.bluestemsoftware.open.eoa.ext.binding.soap.dfault.util.Constants;
import org.bluestemsoftware.open.eoa.ext.binding.soap.dfault.util.FileUtils;
import org.bluestemsoftware.open.eoa.ext.binding.soap.dfault.util.ResponseProcessor;
import org.bluestemsoftware.open.eoa.ext.binding.soap.dfault.util.SOAPFaultContext;
import org.bluestemsoftware.open.eoa.ext.binding.soap.dfault.util.ResponseProcessor.Work;
import org.bluestemsoftware.open.eoa.ext.binding.soap.dfault.util.jms.JMSSender;
import org.bluestemsoftware.open.eoa.ext.binding.soap.dfault.wsa.DuplicateMessageIDFaultImpl;
import org.bluestemsoftware.open.eoa.ext.binding.soap.dfault.wsa.WSAHandler;
import org.bluestemsoftware.open.eoa.ext.binding.soap.dfault.wsse.WSSEHandler;
import org.bluestemsoftware.specification.eoa.component.application.rt.ApplicationRT;
import org.bluestemsoftware.specification.eoa.component.binding.rt.AcceptedResponseError;
import org.bluestemsoftware.specification.eoa.component.binding.rt.BindingFaultRT;
import org.bluestemsoftware.specification.eoa.component.engine.rt.BindingModule;
import org.bluestemsoftware.specification.eoa.component.engine.rt.EndpointActionReference;
import org.bluestemsoftware.specification.eoa.component.engine.rt.EndpointOperationReference;
import org.bluestemsoftware.specification.eoa.component.engine.rt.EndpointReference;
import org.bluestemsoftware.specification.eoa.component.engine.rt.EngineRT;
import org.bluestemsoftware.specification.eoa.component.engine.rt.FeatureModule.WSSpecification;
import org.bluestemsoftware.specification.eoa.component.intrface.InterfaceAction;
import org.bluestemsoftware.specification.eoa.component.intrface.InterfaceFaultReference;
import org.bluestemsoftware.specification.eoa.component.intrface.InterfaceAction.Direction;
import org.bluestemsoftware.specification.eoa.component.intrface.InterfaceOperation.MEP;
import org.bluestemsoftware.specification.eoa.component.intrface.rt.ActionContext;
import org.bluestemsoftware.specification.eoa.component.intrface.rt.FaultContext;
import org.bluestemsoftware.specification.eoa.component.intrface.rt.MessageContext;
import org.bluestemsoftware.specification.eoa.component.intrface.rt.MessageModule;
import org.bluestemsoftware.specification.eoa.component.intrface.rt.SystemFault;
import org.bluestemsoftware.specification.eoa.component.intrface.rt.ActionContext.MessageModules;
import org.bluestemsoftware.specification.eoa.component.message.rt.Content;
import org.bluestemsoftware.specification.eoa.ext.binding.soap.SOAPBindingException;
import org.bluestemsoftware.specification.eoa.ext.binding.soap.rt.SOAPApplicationFault;
import org.bluestemsoftware.specification.eoa.ext.binding.soap.rt.SOAPBindingConfiguration;
import org.bluestemsoftware.specification.eoa.ext.binding.soap.rt.SOAPFault;
import org.bluestemsoftware.specification.eoa.ext.binding.soap.rt.SOAPReceiverFault;
import org.bluestemsoftware.specification.eoa.ext.feature.ws.addressing.WSA;
import org.bluestemsoftware.specification.eoa.ext.feature.ws.addressing.WSAMessageModule;
import org.bluestemsoftware.specification.eoa.ext.feature.ws.addressing.WSA.WSA10;
import org.bluestemsoftware.specification.eoa.ext.feature.ws.addressing.soap.WSABindingModule;
import org.bluestemsoftware.specification.eoa.ext.feature.ws.addressing.soap.fault.WSAddressingFault;
import org.bluestemsoftware.specification.eoa.ext.feature.ws.appdata.ApplicationData;
import org.bluestemsoftware.specification.eoa.ext.feature.ws.appdata.ApplicationDataModule;
import org.bluestemsoftware.specification.eoa.ext.feature.ws.security.WSSEMessageModule;
import org.bluestemsoftware.specification.eoa.ext.feature.ws.security.WSSE.WSSE11;
import org.bluestemsoftware.specification.eoa.ext.feature.ws.transport.vm.VMTransportModule;
import org.bluestemsoftware.specification.eoa.ext.feature.ws.transport.vm.VMTransportRequest;
import org.bluestemsoftware.specification.eoa.system.SystemContext;
import org.bluestemsoftware.specification.eoa.system.System.Log;
import org.w3c.dom.Document;
import org.w3c.dom.Element;

public abstract class SOAPVMTPBindingImpl extends SOAPBindingImpl {

    private static final Log log = SystemContext.getContext().getSystem().getLog(SOAPVMTPBindingImpl.class);

    public SOAPVMTPBindingImpl(SOAPBindingConfiguration configuration, Set<String> features) {
        super(configuration, features);
    }

    @Override
    public void spi_init() throws SOAPBindingException {
        super.spi_init();
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.ext.binding.soap.rt.SOAPVMBindingRT$Provider#spi_receiveAction(org.bluestemsoftware.specification.eoa.component.engine.rt.EndpointReference,
     *      org.w3c.dom.Element)
     */
    public void spi_receiveAction(EndpointReference epr, Direction direction, Element soapEnvelope) {

        log.trace("receiveAction begin");
        
        // because we do not have to parse envelope from a remote stream, we
        // do not have to worry about media types, content encoding, malformed
        // requests, etc ..., i.e., we have, according to SOAP MEP state model,
        // successfully transitioned from the 'init' state to the 'receiving'
        // state, i.e. the message has been 'accepted for processing.'. note
        // that only one provider per binding type can exist within context of
        // an eoa system which mean an instance of this class processed the
        // envelope. so ... we can safely cast to our impl

        if (direction == Direction.IN) {
            receiveMyServiceRequest(epr, (SOAPEnvelope)soapEnvelope);
        } else {
            receivePartnerServiceResponse(epr, (SOAPEnvelope)soapEnvelope);
        }

        log.trace("receiveAction end");

    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.ext.binding.soap.rt.SOAPVMBindingRT$Provider#spi_sendAction(org.bluestemsoftware.specification.eoa.component.engine.rt.EndpointReference,
     *      org.bluestemsoftware.specification.eoa.component.intrface.rt.ActionContext)
     */
    public void spi_sendAction(EndpointReference epr, ActionContext actionContext) throws SystemFault {
        log.debug("spi_sendAction begin");

        ApplicationRT app = epr.getRootComponent().getApplication();
        EndpointActionReference ear = epr.getEndpointActionReference(actionContext.getAction());
        
        Direction direction = null;
        if (ear == null) {
            if (actionContext.getAction().equals(SystemFault.ACTION)) {
                direction = Direction.OUT;
            } else {
                throw new SystemFault(app, "Action '"
                        + actionContext.getAction()
                        + "' is undefined on referenced endpoint "
                        + epr.getEndpoint().getFragmentIdentifier());
            }
        } else {
            direction = ear.getReferencedComponent().getInterfaceAction().getDirection();
        }
        
        try {
        
            if (direction == Direction.OUT) {
                sendMyServiceResponse(epr, actionContext);
            } else {
                sendPartnerServiceRequest(epr, (MessageContext)actionContext);
            }
            
        } catch (Throwable th) {

            // if we end up here, the implication is that message was not accepted
            // by receiver. we translate into a system fault and throw back to
            // engine and allow it to handle it

            SystemFault systemFault = null;

            if (th instanceof SystemFault) {
                if (log.isDebugEnabled()) {
                    log.debug("caught system fault while sending action "
                            + actionContext.getAction()
                            + " defined on epr "
                            + epr.getFragmentIdentifier()
                            + ". "
                            + ((SystemFault)th).getFaultReason());
                }
                systemFault = (SystemFault)th;
            } else if (th instanceof BindingFaultRT) {
                if (log.isDebugEnabled()) {
                    log.debug("caught binding fault while sending action "
                            + actionContext.getAction()
                            + " defined on epr "
                            + epr.getFragmentIdentifier()
                            + ". "
                            + ((BindingFaultRT)th).getFaultReason());
                }
                QName faultCode = ((BindingFaultRT)th).getFaultName();
                systemFault = new SystemFault(app, faultCode, ((BindingFaultRT)th).getFaultReason());
            } else {
                log.error("caught throwable while sending action "
                        + actionContext.getAction()
                        + " defined on epr "
                        + epr.getFragmentIdentifier()
                        + ". "
                        + th);
                systemFault = new SystemFault(app, "Caught throwable sending action.", th);
            }

            throw systemFault;

        }

        log.trace("sendAction end");
    }

    /*
     * action represents a request addressed to a service hosted by target engine. message
     * direction is in and all supported meps are valid
     */
    private void receiveMyServiceRequest(EndpointReference epr, SOAPEnvelope requestEnvelope) {

        log.trace("receiveMyServiceRequest begin");

        // note that invocation of this method implies that we have 'accepted'
        // the request for processing

        MessageModules requestModules = null;
        WSAMessageModule wsammIn = null;
        EndpointActionReference ear = null;
        EndpointOperationReference eor = null;
        MEP mep = null;

        try {
            
            log.debug("receiving my service request");

            // engage ws-addressing. note that we do not persist message
            // record

            WSAHandler wsaHandler = WSAHandler.getInstance(requestEnvelope);
            wsammIn = wsaHandler.handleMyServiceRequest(requestEnvelope);
            requestModules = new MessageModules(wsammIn);
            
            // attempt to register the message. if id is not unique, a sql
            // exception is thrown, i.e. an index constraint violation

            try {
                messageDAO.insertMessage(wsammIn.getMessageID(), true);
            } catch (SQLIntegrityConstraintViolationException se) {
                throw new DuplicateMessageIDFaultImpl(getSOAPVersion(), wsammIn);
            }

            // note that the action is guaranteed to be declared, i.e. the
            // sendPartnerServiceRequest method performs the check and
            // wsa:To is guaranteed to match address used to send request,
            // i.e. sendPartnerServiceRequest overwrites any value which
            // may have been set by user on request context

            String action = requestModules.getAction();
            ear = epr.getEndpointActionReference(action);
            eor = (EndpointOperationReference)ear.getParent();
            mep = eor.getEndpointOperation().getInterfaceOperation().getMessageExchangePattern();

            // process ws-security module first, which may have to decrypt all or
            // part of message (which would be a user configuration error, i.e.
            // message should not be encrypted when transported within jvm)

            WSSEHandler wsseh = WSSEHandler.getInstance(WSSE11.NAMESPACE_URI);
            WSSEMessageModule wsseIn = wsseh.handleMyServiceRequest(epr, action, requestEnvelope);

            if (wsseIn != null) {
                requestModules.putModule(wsseIn);
            }

            // iterate over declared soap headers and verify that, if required,
            // the header is defined on message. ws-a headers have already been
            // extracted and are therefore assumed to exist

            checkRequiredSOAPHeaders(getConcreteMessageDefinition(wsammIn.getAction()), requestEnvelope);

            // execute binding modules. note that modules natively implemented by
            // this binding are a no-op

            MessageModule mm = null;
            for (BindingModule bm : ear.getBindingModules()) {
                mm = bm.handleMyServiceRequest(requestModules, (Content)requestEnvelope);
                if (mm != null) {
                    requestModules.putModule(mm);
                }
            }

            // process remaining headers with application data handler. it
            // will extract headers targeted to ultimate destination including
            // any with role matching role implemented by referenced service.
            // if a header is flagged as mustUnderstand, the handler will
            // exclude it from module unless it was defined via wsad policy

            Map<QName, Set<ApplicationData>> ad = readApplicationData(ear, requestEnvelope);
            if (ad != null) {
                requestModules.getModule(ApplicationDataModule.class).setApplicationData(ad);
            }

            // iterate over remaining headers and throw a must understand
            // fault if any headers remain that we don't understand

            checkMustUnderstand(requestEnvelope);

            // serialize message content to a file to avoid buffering entire
            // message to memory which is limitation of jms api. we use
            // message id as file name, which will be used by jms receiver
            // to retrieve from file system

            String fileName = FileUtils.createFileName(wsammIn.getMessageID());
            File requestContent = new File(dataDirectory, fileName);

            // if payload is null, i.e. it's an empty request. we create an
            // empty file, i.e. which will allow user to manage the request
            // files within the binding var dir, i.e. browsing/deleting,
            // etc ... if jms receiver can't find file, it quietly logs a
            // debug message

            OMElement payload = requestEnvelope.getBody().getFirstElement();

            if (payload != null) {
                OutputStream out = null;
                XMLStreamWriter writer = null;
                try {
                    out = new FileOutputStream(requestContent);
                    writer = STAXUtils.createXMLStreamWriter(out, Constants.UTF_8);
                    if (payload != null) {
                        payload.buildNext(); // kids won't serialize otherwise
                        payload.serializeAndConsume(writer);
                    }
                } finally {
                    if (out != null) {
                        out.flush();
                        IOUtils.closeQuietly(out);
                    }
                    if (writer != null) {
                        writer.close();
                    }
                }
            } else {
                requestContent.createNewFile();
            }
            
            // persist message modules to dbms before passing execution off
            // to a separate thread

            for (MessageModule messageModule : requestModules.values()) {
                messageModuleDAO.insertMessageModule(wsammIn.getMessageID(), true, messageModule);
            }

            // write message to queue where it will be picked up by jms
            // receiver on separate thread and processed by read message
            // method on super
            
            log.debug("writing my service request to queue");

            JMSSender jmsSender = null;
            try {
                jmsSender = (JMSSender)jmsSenders.borrowObject();
                javax.jms.Message msg = jmsSender.createMessage();
                msg.setJMSCorrelationID(wsammIn.getMessageID());
                msg.setStringProperty(Constants.MESSAGE_ID_PROPERTY, wsammIn.getMessageID());
                msg.setBooleanProperty(Constants.MY_SERVICE_PROPERTY, true);
                jmsSender.send(msg, receiveQueue, 0, DeliveryMode.PERSISTENT);
            } finally {
                jmsSenders.returnObject(jmsSender);
            }

        } catch (Throwable th) {

            SOAPFault sf = null;

            if (th instanceof SOAPFault) {
                sf = (SOAPFault)th;
                if (log.isDebugEnabled()) {
                    log.debug("caught soap fault while receiving request action "
                            + wsammIn.getAction()
                            + " defined on endpoint referenced by component "
                            + epr.getFragmentIdentifier()
                            + ". "
                            + ((SOAPFault)th).getFaultReason());
                }
            } else {
                if (configuration.isEmbeddStackTraceInFaults()) {
                    sf = new SOAPReceiverFault(getSOAPVersion(), th);
                } else {
                    sf = new SOAPReceiverFault(getSOAPVersion(), th.getMessage());
                }
                log.error("caught throwable while receiving request action "
                        + wsammIn.getAction()
                        + " defined on endpoint referenced by component "
                        + epr.getFragmentIdentifier()
                        + ". "
                        + th);
            }

            ear = epr.getEndpointActionReference(wsammIn.getAction());

            if (ear != null) {
                eor = (EndpointOperationReference)ear.getParent();
                mep = eor.getEndpointOperation().getInterfaceOperation().getMessageExchangePattern();
            }

            if (mep != null && mep == MEP.IN_ONLY) {
                log.debug("MEP is in-only. No fault response allowed.");
                return;
            }

            if (wsammIn.getMessageID().equals(WSA10.UNSPECIFIED_MESSAGE)) {
                log.debug("unable to return wsa fault response async. no message id.");
                return;
            }

            // process the asynchronous fault response on a separate thread, i.e. so
            // that we can 'accept request' before returning response

            ResponseProcessor responseProcessor = null;
            try {
                responseProcessor = (ResponseProcessor)responseProcessorPool.borrowObject();
                responseProcessor.setWork(new WorkImpl(sf, epr, requestModules));
            } catch (Exception ex) {
                log.error("caught throwable returning async binding fault. " + ex);
            } finally {
                try {
                    responseProcessorPool.returnObject(responseProcessor);
                } catch (Exception fatchance) {
                }
            }

            return;

        }

        log.trace("receiveMyServiceRequest end");

    }

    /*
     * action represents a response from a partner service referenced by target engine. message
     * direction is out and mep is in-out or robust in-only
     */
    private void receivePartnerServiceResponse(EndpointReference epr, SOAPEnvelope responseEnvelope) {

        log.trace("receivePartnerServiceResponse begin");

        // invocation of this method implies message has been accepted by receiver

        ApplicationRT app = ((EngineRT)epr.getRootComponent()).getApplication();

        MessageModules responseModules = null;
        MessageModules requestModules = null;
        WSAMessageModule wsamm = null;
        List<Content> content = null;

        try {
            
            log.debug("receiving partner service response");

            // attempt to engage ws-addressing. if this fails, a soap fault
            // is thrown by handler. if not engaged, we cannot continue

            String action = null;
            WSAHandler wsaHandler = WSAHandler.getInstance(responseEnvelope);
            try {
                wsamm = wsaHandler.handlePartnerServiceResponse(responseEnvelope);
                responseModules = new MessageModules(wsamm);
                action = wsamm.getAction();
            } catch (WSAddressingFault wsaf) {
                if (log.isDebugEnabled()) {
                    log.debug("unable to process partner response from epr "
                            + epr.getFragmentIdentifier()
                            + ". failed to engage ws-addressing. "
                            + wsaf.getFaultReason());
                }
                return;
            }

            // using relatesTo message id defined on wsa header, look up and
            // attempt to correllate with an outstanding request. if we fail
            // to correllate, we cannot continue

            String requestMessageID = wsamm.getRelatesTo();
            requestModules = messageModuleDAO.selectMessageModules(requestMessageID, false);

            if (requestModules == null) {
                log.warn("unable to process partner response from epr "
                        + epr.getFragmentIdentifier()
                        + ". failed to correllate relatesTo '"
                        + requestMessageID
                        + "' with a request.");
                return;
            }
            
            // attempt to register the message. if id is not unique, a sql
            // exception is thrown, i.e. an index constraint violation.
            // note that if partner failed to supply a messageid, one was
            // generated by handler

            try {

                messageDAO.insertMessage(wsamm.getMessageID(), false);

            } catch (SQLIntegrityConstraintViolationException se) {

                // fault handling logic depends upon the existence of
                // a message id/record. so, log an error and create a
                // new module with a unique id. a partner supplied msg
                // id on a response is more or less irrelevant and
                // non-standard and therefore shouldn't be a problem

                WSAMessageModule temp = new WSAMessageModule(action);
                temp.setTo(wsamm.getTo());
                temp.setNamespace(wsamm.getNamespace());
                temp.setRelatesTo(wsamm.getMessageID());

                log.debug("partner supplied message id '"
                        + wsamm.getMessageID()
                        + "' defined on response action "
                        + wsamm.getAction()
                        + " is not unique. "
                        + " overwriting with unique id '"
                        + temp.getMessageID()
                        + "'.");

                wsamm = temp;

                messageDAO.insertMessage(wsamm.getMessageID(), false);

            }

            // process ws-security module first, which may have to decrypt all or
            // part of message which would, in this case, be a user configuration
            // error, i.e. message should not be encrypted when transported within
            // vm. if security handler fails, we cannot continue

            WSSEHandler wsseh = WSSEHandler.getInstance(WSSE11.NAMESPACE_URI);
            WSSEMessageModule wssemm = null;
            try {
                wssemm = wsseh.handlePartnerServiceResponse(epr, action, responseEnvelope);
            } catch (SOAPFault sf) {
                if (log.isDebugEnabled()) {
                    log.debug("unable to process partner response from epr "
                            + epr.getFragmentIdentifier()
                            + " due to a ws-security related error. "
                            + sf.getFaultReason());
                }
                return;
            }

            if (wssemm != null) {
                responseModules.putModule(wssemm);
            }

            // retrieve message payload which may be null. note that if
            // it's a fault response, payload is extracted from detail
            // element of fault

            QName faultCode = null;
            if (responseEnvelope.getBody().hasFault()) {
                faultCode = SOAPFault.getFaultName((Element)responseEnvelope);
                content = SOAPFault.getFaultDetail((Element)responseEnvelope);
            } else {
                Content temp = (Content)responseEnvelope.getBody().getFirstElement();
                if (temp != null) {
                    content = Arrays.asList(temp);
                }
            }

            // iterate over declared soap headers and verify that, if required,
            // the header is defined on message. ws-a headers have already been
            // extracted and are therefore assumed to exist

            checkRequiredSOAPHeaders(getConcreteMessageDefinition(action), responseEnvelope);

            EndpointActionReference ear = epr.getEndpointActionReference(action);
            List<BindingModule> bindingModules = null;
            if (ear == null) {
                bindingModules = epr.getDefaultBindingModules();
            } else {
                bindingModules = ear.getBindingModules();
            }

            // iterate over binding modules and invoke each. note that modules natively
            // implemented by this binding are a no-op

            for (BindingModule bindingModule : bindingModules) {
                Content msg = (Content)responseEnvelope;
                MessageModule rm = null;
                try {
                    rm = bindingModule.handlePartnerServiceResponse(requestModules, responseModules, msg);
                } catch (Throwable th) {
                    String faultReason = null;
                    if (th instanceof SOAPFault) {
                        faultCode = ((SOAPFault)th).getFaultName();
                        faultReason = ((SOAPFault)th).getFaultReason();
                        if (log.isDebugEnabled()) {
                            log.debug("binding module threw a binding fault. " + faultReason);
                        }
                    } else {
                        faultCode = BindingFaultRT.RECEIVER_ERROR;
                        faultReason = th.getMessage();
                        log.error("binding module threw an unexpected exception. " + th);
                    }
                    throw new AcceptedResponseError(faultCode, faultReason, content);
                }
                if (rm != null) {
                    responseModules.putModule(rm);
                }
            }

            // process remaining headers with application data handler. it
            // will extract headers targeted to ultimate destination including
            // any with role matching role implemented by referenced service.
            // if a header is flagged as mustUnderstand, the handler will
            // exclude it from module unless it was defined via wsad policy

            if (ear != null) {
                Map<QName, Set<ApplicationData>> ad = readApplicationData(ear, responseEnvelope);
                if (ad != null) {
                    responseModules.getModule(ApplicationDataModule.class).setApplicationData(ad);
                }
            }

            // iterate over remaining headers and throw a must understand
            // fault if any found that we don't understand

            checkMustUnderstand(responseEnvelope);

            // now we must handle undeclared actions which is either an
            // undeclared normal response or an undeclared fault action

            if (ear == null) {
                String faultReason = null;
                if (faultCode != null) {
                    faultReason = SOAPFault.getFaultReason((Element)responseEnvelope);
                } else {
                    faultCode = BindingFaultRT.RECEIVER_ERROR;
                    faultReason = "Response action "
                            + wsamm.getAction()
                            + " defined on partner response is undeclared.";
                }
                throw new AcceptedResponseError(faultCode, faultReason, content);
            }

            // so if we make it to this point, we have a declared action which
            // maps to a payload. note that payload could be a normal message,
            // a fault message, or empty. payload will be verified by our soap
            // reader when pulled of queue by super

        } catch (Throwable th) {

            // because response has been accepted for processing, any exceptions
            // caught here must be communicated to engine via receive action method.
            // convert the exception into an accepted response error which will
            // be used to generate an instance of system fault

            AcceptedResponseError are = null;

            if (th instanceof AcceptedResponseError) {
                are = (AcceptedResponseError)th;
            } else {
                QName faultCode = null;
                String faultReason = null;
                if (th instanceof SOAPFault) {
                    faultCode = ((SOAPFault)th).getFaultName();
                    faultReason = ((SOAPFault)th).getFaultReason();
                    if (log.isDebugEnabled()) {
                        log.debug("caught soap fault while processing partner response. " + faultReason);
                    }
                } else {
                    faultReason = "Caught throwable while processing partner response. " + th.getMessage();
                    log.error("caught throwable while processing partner response. " + th);
                }
                are = new AcceptedResponseError(faultCode, faultReason, content);
            }

            // because partner is an eoa compliant engine, they may have thrown a
            // system fault, in which case, use existing payload, which should
            // be a system fault message

            boolean useExistingPayload = false;

            if (wsamm.getAction().equals(SystemFault.ACTION)) {
                useExistingPayload = true;
            }

            // translate soap fault into a system fault so that we can use it
            // to generate an instanceof system fault message which contains
            // partner response within fault detail

            if (useExistingPayload == false) {
                SystemFault systemFault = SystemFault.valueOf(app, are);
                Content temp = systemFault.toMessage().getContent();
                Document owner = ((Element)responseEnvelope).getOwnerDocument();
                content = Arrays.asList((Content)owner.importNode(temp, true));
            }

            // payload is a runtime instance of system fault message.
            // change action to system fault action, i.e. engine
            // should never see an undeclared action uri

            wsamm.setAction(SystemFault.ACTION);

        }
        
        // persist message modules to dbms before passing execution off
        // to a separate thread

        for (MessageModule messageModule : responseModules.values()) {
            try {
                messageModuleDAO.insertMessageModule(wsamm.getMessageID(), false, messageModule);
            } catch (Exception ex) {
                log.error("error persisting message modules. " + ex);
                return;
            }
        }

        // write message to queue where it will be picked up by jms
        // receiver on separate thread and processed by read message
        // method on super

        JMSSender jmsSender = null;
        try {
            
            log.debug("writing partner response to queue");

            // serialize message content to a file to avoid buffering entire
            // message to memory as required by jms api. we use message id as
            // file name, which will be used by jms receiver to retrieve from
            // file system

            String fileName = FileUtils.createFileName(wsamm.getMessageID());
            File responseContent = new File(dataDirectory, fileName);

            // if payload is null, i.e. it's an empty response. we create an
            // empty file, i.e. which will allow user to manage the request
            // files within the binding var dir, i.e. browsing/deleting,
            // etc ... if jms receiver can't find file, it quietly logs a
            // debug message

            if (content != null) {
                OutputStream out = null;
                XMLStreamWriter writer = null;
                try {
                    out = new FileOutputStream(responseContent);
                    writer = STAXUtils.createXMLStreamWriter(out, Constants.UTF_8);
                    OMElement temp = (OMElement)content.iterator().next();
                    temp.buildNext(); // kids won't serialize otherwise
                    temp.serializeAndConsume(writer);
                } finally {
                    if (out != null) {
                        out.flush();
                        IOUtils.closeQuietly(out);
                    }
                    if (writer != null) {
                        writer.close();
                    }
                }
            } else {
                responseContent.createNewFile();
            }

            jmsSender = (JMSSender)jmsSenders.borrowObject();
            javax.jms.Message msg = jmsSender.createMessage();
            msg.setStringProperty(Constants.MESSAGE_ID_PROPERTY, wsamm.getMessageID());
            msg.setBooleanProperty(Constants.MY_SERVICE_PROPERTY, false);
            jmsSender.send(msg, receiveQueue, 0, DeliveryMode.PERSISTENT);

        } catch (Throwable th) {
            log.error("caught throwable while receiving partner response. " + th);
        } finally {
            try {
                jmsSenders.returnObject(jmsSender);
            } catch (Exception fatchance) {
            }
        }

        log.trace("receivePartnerServiceResponse end");

    }

    /*
     * action represents a response being returned by a locally hosted service. message
     * direction is out and mep is in-out or robust in-only
     */
    protected void sendMyServiceResponse(EndpointReference epr, ActionContext actionContext) throws Exception {

        log.trace("sendMyServiceResponse begin");

        // if action context is an instanceof SOAPFaultContext, then request
        // failed to enter application layer, in which case request modules
        // are passed by reference. otherwise look-up using relates to value
        
        log.debug("sending my service response");

        ApplicationRT app = ((EngineRT)epr.getRootComponent()).getApplication();
        WSAMessageModule wsammOut = actionContext.getMessageModule(WSAMessageModule.class);
        String requestMessageID = wsammOut.getRelatesTo();
        MessageModules requestModules = null;

        WSAMessageModule wsammIn = null;
        if (actionContext instanceof SOAPFaultContext) {
            requestModules = ((SOAPFaultContext)actionContext).getRequestModules();
        } else {
            requestModules = messageModuleDAO.selectMessageModules(requestMessageID, true);
            if (requestModules == null) {
                throw new SystemFault(app, "Failed to correllate request.");
            }
        }

        wsammIn = requestModules.getModule(WSAMessageModule.class);

        // attempt to persist the message. if id is not unique, a sql
        // exception is thrown

        try {
            messageDAO.insertMessage(wsammOut.getMessageID(), true);
        } catch (SQLIntegrityConstraintViolationException se) {
            throw new SystemFault(app, "Response messageID not unique.");
        }

        SOAPEnvelope responseEnvelope = null;
        String responseAction = actionContext.getAction();
        List<BindingModule> bindingModules;

        EndpointActionReference responseEAR = epr.getEndpointActionReference(responseAction);

        boolean isFaultResponse = false;

        if (responseEAR != null) {

            EndpointOperationReference eor = (EndpointOperationReference)responseEAR.getParent();
            MEP mep = eor.getEndpointOperation().getInterfaceOperation().getMessageExchangePattern();

            if (mep == MEP.IN_ONLY) {
                throw new SystemFault(app, "MEP is 'in-only'. No response is allowed.");
            }

            InterfaceAction ia = responseEAR.getReferencedComponent().getInterfaceAction();
            if (ia instanceof InterfaceFaultReference) {
                SOAPApplicationFault sf = new SOAPApplicationFault(getSOAPVersion(), (FaultContext)actionContext);
                responseEnvelope = buildSOAPFaultEnvelope(epr, sf);
                isFaultResponse = true;
            } else {
                if (mep == MEP.ROBUST_IN_ONLY) {
                    throw new SystemFault(app, "MEP is 'robust-in-only'. Normal response not allowed.");
                }
                responseEnvelope = writeMessage(epr, (MessageContext)actionContext);
            }

            bindingModules = responseEAR.getBindingModules();

        } else {

            // the action is undeclared. we run through default set of binding
            // modules so that required response headers will be generated

            if (actionContext instanceof SOAPFaultContext) {

                // a soap fault was generated before it made it into application
                // layer. no need to bind message just convert to soap fault
                // envelope

                SOAPFault soapFault = ((SOAPFaultContext)actionContext).getSoapFault();
                responseEnvelope = buildSOAPFaultEnvelope(epr, soapFault);

            } else {

                // the only undeclared action we support is the system fault
                // action. throw it back to engine if that's not the case

                if (!actionContext.getAction().equals(SystemFault.ACTION)) {
                    throw new SystemFault(app, "Action " + actionContext.getAction() + " is undefined.");
                }

                // wrap system fault in a soap application fault object which
                // will handle manipulation of fault code, detail and action

                SOAPApplicationFault sf = new SOAPApplicationFault(getSOAPVersion(), (FaultContext)actionContext);
                responseEnvelope = buildSOAPFaultEnvelope(epr, sf);

            }

            isFaultResponse = true;
            bindingModules = epr.getDefaultBindingModules();

        }

        // execute natively implemented binding modules. the version of spec
        // employed is version selected by client, i.e. implied by namespace
        // of request modules

        WSAHandler wsah = WSAHandler.getInstance(wsammIn.getNamespace());
        MessageModules responseModules = actionContext.getMessageModules();
        wsammOut = wsah.handleMyServiceResponse(requestModules, responseModules, responseEnvelope);
        responseModules.putModule(wsammOut);

        // execute binding modules. note that modules natively implemented by
        // this binding are a no-op

        for (BindingModule bm : bindingModules) {
            MessageModule mm = bm.handleMyServiceResponse(requestModules, (Content)responseEnvelope);
            if (mm != null) {
                responseModules.putModule(mm);
            }
        }

        // write all application data passed to us by application as soap
        // header blocks

        if (responseEAR != null) {
            writeApplicationData(responseEAR, responseModules, responseEnvelope);
        }

        // process ws-security module last, which may encrypt and sign all or
        // part of message. note that currently we support only one version of
        // ws-se, but spec allows support of future versions of wsa spec

        WSSEMessageModule wssemOut = null;
        WSSEMessageModule wssemIn = requestModules.getModule(WSSEMessageModule.class);
        if (wssemIn != null) {
            WSSEHandler wsseh = WSSEHandler.getInstance(wssemIn.getNamespace());
            wssemOut = wsseh.handleMyServiceResponse(requestModules, responseModules, responseEnvelope);
            if (wssemOut != null) {
                responseModules.putModule(wssemOut);
            }
        }

        // persist the response modules to dbms
        for (MessageModule mm : responseModules.values()) {
            messageModuleDAO.insertMessageModule(wsammOut.getMessageID(), true, mm);
        }

        // determine replyTo according to whether or not this is a fault,
        // and whether or not a faultTo address exists

        String replyTo = null;
        if (isFaultResponse) {
            if (wsammIn.getFaultTo() == null) {
                replyTo = wsammIn.getReplyTo().getAddress();
            } else {
                replyTo = wsammIn.getFaultTo().getAddress();
            }
        } else {
            replyTo = wsammIn.getReplyTo().getAddress();
        }

        // return response by transporting within vm
        
        log.debug("returning my service response via vm transport module");
        VMTransportModule tm = (VMTransportModule)epr.getDefaultTransportModule();
        tm.doSend(new VMTransportRequest(replyTo, Direction.OUT, (Element)responseEnvelope));

        log.trace("sendMyServiceResponse end");

    }

    /*
     * action represents a request addressed to a partner service. message direction is in and
     * all supported meps are valid
     */
    private void sendPartnerServiceRequest(EndpointReference epr, MessageContext messageContext) throws SystemFault {

        log.trace("sendPartnerServiceRequest begin");

        // as required by spec, if an error occurs before we transition to
        // the 'sending+receiving' state, we throw a fault to spi_sendAction
        // method where it will be converted into a system fault and thrown
        // back to engine as an indication action was not accepted by partner
        
        log.debug("sending partner service request");

        ApplicationRT app = ((EngineRT)epr.getRootComponent()).getApplication();
        MessageModules requestModules = messageContext.getMessageModules();
        WSAMessageModule wsammIn = requestModules.getModule(WSAMessageModule.class);

        EndpointActionReference requestAction = epr.getEndpointActionReference(wsammIn.getAction());
  
        // attempt to persist the message. if id is not unique, a sql
        // exception is thrown

        try {
            messageDAO.insertMessage(wsammIn.getMessageID(), false);
        } catch (SQLIntegrityConstraintViolationException sve) {
            throw new SystemFault(app, "Request messageID " + wsammIn.getMessageID() + " not unique.");
        } catch (SQLException se) {
            throw new SystemFault(app, "Error persisting request. " + se);
        }

        // convert abstract message instance along with any application data
        // into a version specific soap envelope which represents concrete
        // message instance

        SOAPEnvelope requestEnvelope = writeMessage(epr, (MessageContext)messageContext);

        // currently we support only one version of wsa, but spec allows
        // support for future versions of wsa spec

        WSABindingModule wsabm = requestAction.getBindingModule(WSABindingModule.class);
        WSA wsa = (WSA)wsabm.getSupportedVersions().values().iterator().next();
        WSSpecification.Policy wsaPolicy = wsa.getPolicy().iterator().next();
        Boolean isAnonymousResponses = ((WSA10.Policy)wsaPolicy).isAnonymousResponses();

        EndpointOperationReference eor = (EndpointOperationReference)requestAction.getParent();
        MEP mep = eor.getEndpointOperation().getInterfaceOperation().getMessageExchangePattern();

        // set response epr(s). if policy is non specific regarding anonymous
        // responses, check for a replyTo address set via private policy. note
        // that ws-transport-vm verified that policy requires non-anonymous
        // replies, but we check again, just in case

        org.bluestemsoftware.specification.wsa.EndpointReference replyTo = null;
        org.bluestemsoftware.specification.wsa.EndpointReference faultTo = null;

        if (mep == MEP.IN_ONLY) {
            replyTo = WSA10.NONE_REPLY_TO;
        } else if (mep == MEP.ROBUST_IN_ONLY) {
            replyTo = WSA10.NONE_REPLY_TO;
            if (isAnonymousResponses == null) {
                String replyToAddress = ((WSA10.Policy)wsaPolicy).getReplyToAddress();
                if (replyToAddress == null) {
                    throw new SystemFault(app, "Anonymous response not supported with VM Transport");
                } else {
                    faultTo = new org.bluestemsoftware.specification.wsa.EndpointReference(replyToAddress);
                }
                isAnonymousResponses = Boolean.FALSE;
            } else if (isAnonymousResponses == false) {
                String replyToAddress = ((WSA10.Policy)wsaPolicy).getReplyToAddress();
                faultTo = new org.bluestemsoftware.specification.wsa.EndpointReference(replyToAddress);
            } else {
                throw new SystemFault(app, "Anonymous response not supported with VM Transport");
            }
        } else if (mep == MEP.IN_OUT) {
            if (isAnonymousResponses == null) {
                String replyToAddress = ((WSA10.Policy)wsaPolicy).getReplyToAddress();
                if (replyToAddress == null) {
                    throw new SystemFault(app, "Anonymous response not supported with VM Transport");
                } else {
                    replyTo = new org.bluestemsoftware.specification.wsa.EndpointReference(replyToAddress);
                    isAnonymousResponses = Boolean.FALSE;
                }
            } else if (isAnonymousResponses == false) {
                String replyToAddress = ((WSA10.Policy)wsaPolicy).getReplyToAddress();
                replyTo = new org.bluestemsoftware.specification.wsa.EndpointReference(replyToAddress);
            } else {
                throw new SystemFault(app, "Anonymous response not supported with VM Transport");
            }
        }

        // prime ws-addressing message module with properties which will be
        // used by handler to write headers to envelope. set wsa namespace
        // on module which will be serialized with module and used to return
        // my service response of local partner to whom we are sending
        // the request

        String to = epr.getEndpoint().getAddress();
        wsammIn.setTo(to);
        wsammIn.setReplyTo(replyTo);
        wsammIn.setFaultTo(faultTo);
        wsammIn.setNamespace(WSA10.NAMESPACE_URI);
        WSAHandler wsah = WSAHandler.getInstance(WSA10.NAMESPACE_URI);
        wsah.handlePartnerServiceRequest(requestModules, requestEnvelope);

        // invoke binding modules. note that modules natively implemented by
        // this binding are a no-op. if an error is encountered, module
        // throws a soap fault, which will be propogated to engine

        for (BindingModule bm : requestAction.getBindingModules()) {
            MessageModule mmIn = null;
            try {
                mmIn = bm.handlePartnerServiceRequest(requestModules, (Content)requestEnvelope);
            } catch (BindingFaultRT bf) {
                throw new SystemFault(app, bf.getFaultName(), bf.getFaultReason());
            }
            if (mmIn != null) {
                requestModules.putModule(mmIn);
            }
        }

        // write all application data passed to us by application as soap
        // header blocks

        writeApplicationData(requestAction, requestModules, requestEnvelope);

        // process ws-security module last, which may encrypt and sign all or
        // part of message, which would be a user configuration error, i.e.
        // no need to encrypt if transporting within vm

        WSSEHandler wsseh = WSSEHandler.getInstance(WSSE11.NAMESPACE_URI);
        WSSEMessageModule wsseOut = wsseh.handlePartnerServiceRequest(requestModules, requestEnvelope);
        if (wsseOut != null) {
            requestModules.putModule(wsseOut);
        }

        // persist the request modules to the dbms
        for (MessageModule mm : requestModules.values()) {
            try {
                messageModuleDAO.insertMessageModule(wsammIn.getMessageID(), false, mm);
            } catch (Exception ex) {
                throw new SystemFault(app, "Error inserting message modules. " + ex);
            }
        }

        // attempt to send the request. if module is unable to send, it throws a
        // transport fault, which we spi_sendAction method propogates back to
        // engine as a system fault

        log.debug("sending partner service request via vm transport module");
        VMTransportModule tm = (VMTransportModule)epr.getDefaultTransportModule();
        tm.doSend(new VMTransportRequest(to, Direction.IN, (Element)requestEnvelope));

        log.trace("sendPartnerServiceRequest end");
    }

    public class WorkImpl implements Work {

        private SOAPFault sf;
        private EndpointReference epr;
        private MessageModules requestModules;

        public WorkImpl(SOAPFault sf, EndpointReference epr, MessageModules requestModules) {
            this.sf = sf;
            this.epr = epr;
            this.requestModules = requestModules;
        }

        public void doWork() {

            // note that because request was sent by our binding, wsa checks
            // were done within sendPartnerServiceRequest method, i.e. wsa
            // faults are not thrown by receiveMyServiceRequest. wsa is
            // guaranteed to be engaged

            try {
                WSAMessageModule wsammOut = new WSAMessageModule(sf.getAction());
                WSAMessageModule wsammIn = requestModules.getModule(WSAMessageModule.class);
                wsammOut.setRelatesTo(wsammIn.getMessageID());
                MessageModules responseModules = new MessageModules(wsammOut);
                SOAPFaultContext sfc = new SOAPFaultContext(null, requestModules, responseModules, sf);
                sendMyServiceResponse(epr, sfc);
            } catch (Throwable thr) {
                log.error("caught throwable returning async fault response. " + thr);
            }

        }

    }

}
